"""ZOOMA API wrapper."""

import logging

import requests

ZOOMA_API_URL = 'https://www.ebi.ac.uk/spot/zooma/v2/api'

__all__ = [
    'ZoomaClient'
]

logger = logging.getLogger(__name__)


class ZoomaClient:
    """Simple client to query ZOOMA. Will look in all curated datasources and perform a fuzzy search in EFO. Only HIGH
    quality mappings are considered."""

    def __init__(self, zooma_base=ZOOMA_API_URL):
        """
        :param zooma_base: An optional, custom URL for the Zooma RESTful API.
        """
        self.base = zooma_base.rstrip('/')
        self.session = requests.Session()

    def search(self, query_string):
        """Query ZOOMA and return all high confidence mappings."""
        payload = {
            'ontologies': '[EFO]',
        }
        url = f'{self.base}/services/annotate?propertyValue={query_string}'
        for zooma_result in requests.get(url, data=payload).json():
            if zooma_result['confidence'] == 'HIGH':
                for semantic_tag in zooma_result['semanticTags']:
                    yield semantic_tag



import logging
import os
import urllib.request

import pandas as pd
import pronto
import requests
from retry import retry

from ontoma import ontology


EFO_RELEASE_API_TEMPLATE = 'https://api.github.com/repos/EBISPOT/efo/releases/{}'

OWL_FILENAME = 'efo_otar_slim.owl'
TERMS_FILENAME = 'terms.tsv'
XREFS_FILENAME = 'xrefs.tsv'
SYNONYMS_FILENAME = 'synonyms.tsv'
ALL_FILENAMES = {OWL_FILENAME, TERMS_FILENAME, XREFS_FILENAME, SYNONYMS_FILENAME}

logger = logging.getLogger(__name__)


@retry(logger=logger, tries=5, delay=3, backoff=1.5, jitter=(1, 3))
def fetch_efo_assets_from_github(efo_release):
    """Queries GitHub API to fetch the latest EFO release."""
    if efo_release == 'latest':
        url = EFO_RELEASE_API_TEMPLATE.format(efo_release)
    else:
        url = EFO_RELEASE_API_TEMPLATE.format(f'tags/{efo_release}')
    response = requests.get(url)
    response.raise_for_status()  # In case of HTTP errors, this will be caught by the @retry decorator.
    return response.json()['assets']


def preprocess_owl(outdir, efo_release):
    """Downloads and preprocesses the EFO OT slim CWL file. Outputs several tables in TSV format to the specified
    directory."""

    if not os.path.isdir(outdir):
        logging.info('Create the output directory.')
        os.mkdir(outdir)

    if not os.path.isfile(os.path.join(outdir, OWL_FILENAME)):
        logging.info('Download the EFO OT slim OWL file.')
        efo_assets = fetch_efo_assets_from_github(efo_release)
        otar_slim_assets = [a for a in efo_assets if a['name'] == OWL_FILENAME]
        if len(otar_slim_assets) == 0:
            raise AssertionError(f'EFO release {efo_release!r} on GitHub does not contain the file {OWL_FILENAME!r}.')
        if len(otar_slim_assets) > 1:
            raise AssertionError(f'EFO release {efo_release!r} contains multiple files named {OWL_FILENAME!r}.')
        otar_slim_url = otar_slim_assets[0]['browser_download_url']
        urllib.request.urlretrieve(otar_slim_url, os.path.join(outdir, OWL_FILENAME))

    logging.info('Process the OWL file.')
    terms_dataset, xrefs_dataset, synonyms_dataset = [], [], []
    for term in pronto.Ontology(os.path.join(outdir, OWL_FILENAME)).terms():
        normalised_term_id = ontology.normalise_ontology_identifier(term.id)
        terms_dataset.append([
            normalised_term_id,
            term.name.lower() if term.name else '',
            term.obsolete,
        ])
        for xref in term.xrefs:
            normalised_xref_id = ontology.normalise_ontology_identifier(xref.id)
            if normalised_xref_id:
                xrefs_dataset.append([
                    normalised_xref_id,
                    normalised_term_id,
                ])
        for synonym in term.synonyms:
            if synonym.description and synonym.scope == 'EXACT':
                synonyms_dataset.append([
                    synonym.description.lower(),
                    normalised_term_id,
                ])

    logging.info('Output the datasets.')
    pd.DataFrame(
        terms_dataset, columns=('normalised_id', 'normalised_label', 'is_obsolete')
    ).to_csv(os.path.join(outdir, TERMS_FILENAME), sep='\t', index=False)
    pd.DataFrame(
        xrefs_dataset, columns=('normalised_xref_id', 'normalised_id',)
    ).to_csv(os.path.join(outdir, XREFS_FILENAME), sep='\t', index=False)
    pd.DataFrame(
        synonyms_dataset, columns=('normalised_synonym', 'normalised_id')
    ).to_csv(os.path.join(outdir, SYNONYMS_FILENAME), sep='\t', index=False)



"""Main interface class."""

__all__ = [
    'OnToma',
]

from collections import namedtuple
import logging
import os
import tempfile

import pandas as pd

from ontoma import ontology
from ontoma import owl
from ontoma.constants import URLS, RESULT_FIELDS
from ontoma.downloaders import get_manual_xrefs, get_manual_string_mappings
from ontoma.oxo import OxoClient
from ontoma.zooma import ZoomaClient

logger = logging.getLogger(__name__)


OnTomaResult = namedtuple('OnTomaResult', RESULT_FIELDS)


class OnToma:
    """Open Targets ontology mapping wrapper. Please refer to documentation for usage details."""

    def __init__(self, cache_dir=None, efo_release='latest'):
        """Initialise an OnToma instance and fetch the necessary resources. Depending on whether cache_dir is specified
        and whether it contains anything, the following behaviour is applied:
        1. If cache_dir is specified and is not empty, OnToma will use EFO cache from it as is.
        2. If cache_dir is specified and is empty or does not exist, OnToma will download the cache to the directory and
           then use it.
        3. If cache_dir is not specified, a temporary directory will be used to fetch and store EFO cache. Note that
           this cannot be persistent, so the cache would have to be re-downloaded each time.
        """
        self.logger = logger

        # Initialize API clients.
        self._oxo = OxoClient()
        self._zooma = ZoomaClient()

        # Import EFO OWL datasets.
        if not cache_dir:
            cache_dir = tempfile.TemporaryDirectory().name
            self.logger.warning(
                f'EFO cache directory is not specified. Created temporary directory to store the cache: {cache_dir}. '
                f'We recommend that you specify the cache directory to speed up subsequent OnToma runs.'
            )
        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)
            self.logger.info(f'Created EFO cache directory {cache_dir}.')
        if not owl.ALL_FILENAMES <= set(os.listdir(cache_dir)):  # If not all EFO cache files are present.
            self.logger.info(f'Downloading EFO cache to {cache_dir}.')
            owl.preprocess_owl(outdir=cache_dir, efo_release=efo_release)
        self.logger.info(f'Using EFO cache from {cache_dir}.')
        self.efo_terms = pd.read_csv(os.path.join(cache_dir, owl.TERMS_FILENAME), sep='\t')
        self.efo_xrefs = pd.read_csv(os.path.join(cache_dir, owl.XREFS_FILENAME), sep='\t')
        self.efo_synonyms = pd.read_csv(os.path.join(cache_dir, owl.SYNONYMS_FILENAME), sep='\t')
        self.logger.info(f'Loaded {len(self.efo_terms)} terms, {len(self.efo_xrefs)} xrefs, '
                         f'and {len(self.efo_synonyms)} synonyms from EFO cache.')

        # Import manually curated datasets.
        self.manual_xrefs = get_manual_xrefs(URLS['MANUAL_XREF'])
        self.manual_string = get_manual_string_mappings(URLS['MANUAL_STRING'])

    def filter_identifiers_by_efo_current(self, normalised_identifiers):
        """Returns a subset of the idenfitiers which are in EFO and not marked as obsolete."""
        return list(
            self.efo_terms[
                (self.efo_terms.normalised_id.isin(normalised_identifiers)) &
                (~ self.efo_terms.is_obsolete)
            ]
            .normalised_id
        )

    def get_label_from_efo(self, normalised_identifier):
        matches = list(
            self.efo_terms[
                self.efo_terms.normalised_id == normalised_identifier
            ]
            .normalised_label
        )
        assert len(matches) == 1
        return matches[0]

    def step01_owl_identifier_match(self, normalised_identifier):
        """If the term is already present in EFO, return it as is."""
        return self.filter_identifiers_by_efo_current(
            self.efo_terms[self.efo_terms.normalised_id == normalised_identifier].normalised_id
        )

    def step02_owl_db_xref(self, normalised_identifier):
        """If there are terms in EFO referenced by the `hasDbXref` field to the query, return them."""
        return self.filter_identifiers_by_efo_current(
            self.efo_xrefs[self.efo_xrefs.normalised_xref_id == normalised_identifier].normalised_id
        )

    def step03_manual_xref(self, normalised_identifier):
        """Look for the queried term in the manual ontology-to-ontology mapping list."""
        return self.filter_identifiers_by_efo_current(
            self.manual_xrefs[self.manual_xrefs.normalised_xref_id == normalised_identifier].normalised_id
        )

    def step04_oxo_query(self, normalised_identifier):
        """Find cross-references using OxO."""
        oxo_mappings = set()
        for result in self._oxo.search(ids=[normalised_identifier], mapping_target='EFO', distance=2):
            for mapping in result['mappingResponseList']:
                oxo_mappings.add(ontology.normalise_ontology_identifier(mapping['curie']))
        return self.filter_identifiers_by_efo_current(oxo_mappings)

    def step05_owl_name_match(self, normalised_string):
        """Find EFO terms which match the string query exactly."""
        return self.filter_identifiers_by_efo_current(
            self.efo_terms[self.efo_terms.normalised_label == normalised_string].normalised_id
        )

    def step06_owl_exact_synonym(self, normalised_string):
        """Find EFO terms which have the query as an exact synonym."""
        return self.filter_identifiers_by_efo_current(
            self.efo_synonyms[self.efo_synonyms.normalised_synonym == normalised_string].normalised_id
        )

    def step07_manual_mapping(self, normalised_string):
        """Find the query in the manual string-to-ontology mapping database."""
        return self.filter_identifiers_by_efo_current(
            self.manual_string[self.manual_string.normalised_label == normalised_string].normalised_id
        )

    def step08_zooma_high_confidence(self, normalised_string):
        zooma_mappings = {
            ontology.normalise_ontology_identifier(mapping)
            for mapping in self._zooma.search(normalised_string)
        }
        return self.filter_identifiers_by_efo_current(zooma_mappings)

    def step09_owl_related_synonym(self, normalised_string):
        raise NotImplementedError

    def step10_zooma_any(self, normalised_string):
        raise NotImplementedError

    def find_term(
            self,
            query: str,
            code: bool = False,
            suggest: bool = False,
    ) -> list:
        """For a given query (an ontology identifier or a string), find matches in EFO Open Targets slim.

        The algorithm operates in a series of steps. If a given step is successful, the result is returned immediately,
        and the remaining steps are not executed. If all steps fail to provide a match, None is returned. Note that in
        general more than one mapping can be returned. This can happen for complex traits which require more than one
        ontology term to represent them.

        If the `code` flag is specified, it is assumed that the query is an ontology identifier, such as 'OMIM:615632',
        and the following steps are attempted:
        1. See if the term is already in EFO OT slim OWL file.
        2. Match terms by cross-references (hasDbXref) from the OWL file.
        3. Mapping from the manual cross-reference database.
        4. Request through OxO with a distance of 2.

        If the query is a string, the following steps are attempted:
        5. Exact name match from EFO OT slim OWL file.
        6. Exact synonym (hasExactSynonym) from the OWL file.
        7. Mapping from the manual string-to-ontology database.
        8. High confidence mapping from ZOOMA with default parameters.

        The following functionality is planned, but not yet implemented. — If the query is a string, and additionally
        the `suggest` flag is specified, additional steps are attempted:
        9. Inexact synonyms (hasRelatedSynonym) from the OWL file.
        10. Any confidence mapping from ZOOMA with default parameters.

        Args:
            query: Either an ontology identifier, or the disease/phenotype string to be matched to an EFO code.
            code: Whether to treat the query as an ontology identifier.
            suggest: Whether to report low quality mappings which are not guaranteed to be contained in EFO OT slim.

        Returns:
            A list of values dependent on the `verbose` flag (either strings with ontology identifiers, or a dictionary
            of additional information). The list will be empty if no hits were identified."""

        # Attempt mapping using various strategies for identifier/string inputs.
        if code:
            normalised_identifier = ontology.normalise_ontology_identifier(query)
            result = (
                self.step01_owl_identifier_match(normalised_identifier)
                or self.step02_owl_db_xref(normalised_identifier)
                or self.step03_manual_xref(normalised_identifier)
                or self.step04_oxo_query(normalised_identifier)
            )
        else:
            normalised_string = query.lower()
            result = (
                self.step05_owl_name_match(normalised_string)
                or self.step06_owl_exact_synonym(normalised_string)
                or self.step07_manual_mapping(normalised_string)
                or self.step08_zooma_high_confidence(normalised_string)
            )
            if not result and suggest:
                result = (
                        self.step09_owl_related_synonym(normalised_string) +
                        self.step10_zooma_any(normalised_string)
                )

        # Convert the term representation into the format supported by the Open Targets schema.
        processed_results = []
        for r in result:
            processed_results.append(
                OnTomaResult(
                    query=query,
                    id_normalised=r,
                    id_ot_schema=ontology.convert_to_ot_schema(r),
                    id_full_uri=ontology.convert_to_uri(r),
                    label=self.get_label_from_efo(r),
                )
            )

        # Return either the list of dictionaries, or just the mappings, depending on parameters.
        self.logger.info(f'Processed: {query} → {processed_results}')
        return processed_results



"""Contains utilities for handling ontology terms and normalising the identifier representation.

The idea behind this module is that any incoming ontology representation (from query, EFO OWL file etc.) is first
converted into internal, consistent representation. Then direct string-to-string comparisons can be made without having
to adjust for different ontology representation methods. For example, the following identifiers: ORDO_140162,
ORPHA:140162, Orphanet:140162 are all consistently converted into ORDO:140162.

The normalised identifiers can then be converted into other formats for exporting the data or making external queries,
for example:
* The Open Targets schema format: Orphanet_140162;
* Full URI for querying OLS: http://www.orpha.net/ORDO/Orphanet_140162."""

import re
from typing import Optional

# Some ontologies (most notably Orphanet) have multiple labels for specifying the same namespace.
NORMALISATION_MAPPING = {
    'ORPHA': 'ORDO',
    'ORPHANET': 'ORDO',
}


def normalise_ontology_identifier(identifier: str) -> Optional[str]:
    """Normalise ontology identifier representation in order to make direct string-to-string comparison possible."""
    if 'identifiers.org' in identifier:
        # If this is an identifiers.org link, extract ontology name and identifier from the last two elements.
        # http://identifiers.org/omim/137215" → [omim, 137215].
        ontology_name, ontology_id = identifier.split('/')[-2:]
    else:
        # For any other type of link (EFO, OBO, Orphanet), extract ontology name and identifier from the last element.
        # http://www.orpha.net/ORDO/Orphanet_140162 → Orphanet_140162.
        identifier = identifier.split('/')[-1]
        # Split into ontology name and identifier parts, separated by either colon or underscore.
        # Orphanet_140162 → [Orphanet, 140162].
        parts = re.split('[:_]', identifier)
        if len(parts) != 2:  # Not a valid/supportable ontology identifier.
            return None
        ontology_name, ontology_id = parts

    # Convert ontology name to uppercase and substitute, if necessary.
    # [Orphanet, 140162] → [ORDO, 140162].
    ontology_name = ontology_name.upper()
    ontology_name = NORMALISATION_MAPPING.get(ontology_name, ontology_name)

    # Glue ontology name and identifier back together into a CURIE. This can be queried in most contexts and tools.
    # [ORDO, 140162] → ORDO:140162.
    return f'{ontology_name}:{ontology_id}'


OT_SCHEMA_MAPPING = {
    'ORDO': 'Orphanet'
}
OT_SCHEMA_MAPPING.update({
    ontology: ontology for ontology in ('NCIT', 'GO', 'HP', 'EFO', 'MONDO', 'DOID', 'MP')
})
# While OTAR is listed in the schema, it is not included, because it is only used for classes and never for individual
# terms.


def convert_to_ot_schema(normalised_identifier: str) -> Optional[str]:
    """Convert normalised ontology representation into the one accepted by the Open Targets JSON schema. For example,
    ORDO:140162 → Orphanet_140162. If the target ontology is not supported by the schema, return None."""
    ontology_name, ontology_id = normalised_identifier.split(':')
    ontology_name = OT_SCHEMA_MAPPING.get(ontology_name)
    if ontology_name:
        return f'{ontology_name}_{ontology_id}'
    else:
        return None


URI_MAPPING = {
    'NCIT':  'http://purl.obolibrary.org/obo/NCIT_{}',
    'ORDO':  'http://www.orpha.net/ORDO/Orphanet_{}',
    'GO':    'http://purl.obolibrary.org/obo/GO_{}',
    'HP':    'http://purl.obolibrary.org/obo/HP_{}',
    'EFO':   'http://www.ebi.ac.uk/efo/EFO_{}',
    'MONDO': 'http://purl.obolibrary.org/obo/MONDO_{}',
    'DOID':  'http://purl.obolibrary.org/obo/DOID_{}',
    'MP':    'http://purl.obolibrary.org/obo/MP_{}',
}


def convert_to_uri(normalised_identifier: str) -> Optional[str]:
    """Convert normalised ontology representation into a full URI. If the target ontology is not supported by the
    schema, return None."""
    ontology_name, ontology_id = normalised_identifier.split(':')
    uri_template = URI_MAPPING.get(ontology_name)
    if uri_template:
        return uri_template.format(ontology_id)
    else:
        return None



import csv
import logging

import click

from ontoma.constants import RESULT_FIELDS, EFO_DEFAULT_VERSION
from ontoma.interface import OnToma

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    '--infile',
    type=click.File('r'),
    default='/dev/stdin',
    help='File to read the input from, one per line. The type of input is determined by --input-type.'
)
@click.option(
    '--outfile',
    type=click.File('w'),
    default='/dev/stdout',
    help='File to output the mappings to. The format is two column TSV file. First column is your query, and the '
         'second column contains the mappings. When no mappings were found, no lines will be output. When multiple '
         'mappings are found, multiple lines will be output, with one mapping per line.'
)
@click.option(
    '--input-type',
    type=click.Choice(['string', 'ontology']),
    default='string',
    help='Can be one of two values. When set to “string” (default), input values are treated as strings, e.g. '
         '“Diabetes”. When set to “ontology”, input values are treated as ontology identifiers, e.g. “MONDO_0005015”.'
)
@click.option(
    '--cache-dir',
    type=str,
    default=None,
    help='A directory to store EFO cache. Specifying it is not required, but will speed up subsequent OnToma runs.'
)
@click.option(
    '--columns',
    type=str,
    default='query,id_ot_schema',
    help=f'Which columns to output, comma separated. The available options are: {",".join(RESULT_FIELDS)}'
)
@click.option(
    '--efo-release',
    type=str,
    default=EFO_DEFAULT_VERSION,
    help=f'EFO release to use. This must be be either “latest”, or match the specific tag name in their GitHub '
         f'releases, for example v3.31.0. By default, {EFO_DEFAULT_VERSION!r} is used.'
)
def ontoma(infile, outfile, input_type, cache_dir, columns, efo_release):
    """Maps ontology identifiers and strings to EFO, the ontology used by the Open Targets Platform."""
    if infile.name == '/dev/stdin':
        logger.warning('Reading input from STDIN. If this is not what you wanted, re-run with --help to see usage.')

    logger.info('Initialising OnToma main interface.')
    otmap = OnToma(cache_dir=cache_dir, efo_release=efo_release)
    columns = columns.split(',')
    efo_writer = csv.DictWriter(outfile, columns, delimiter='\t')
    efo_writer.writeheader()

    logger.info(f'Treating the input as (type = {input_type}) and mapping to EFO.')
    mapped = 0
    failed = 0
    for line in infile:
        query = line.rstrip()
        results = otmap.find_term(query, code=input_type == 'ontology')
        for result in results:
            efo_writer.writerow({c: getattr(result, c) for c in columns})
        if results:
            mapped += 1
        else:
            failed += 1

    logger.info(f'Processed {mapped + failed} inputs. Of them, found at least one EFO hit for {mapped}, and failed to '
                f'find any hits for {failed}.')
